# function [cD cA] = getDWT(X,N,Name)
#
# finds the discrete wavelet transform at level N for signal X using the
# wavelet specified by Name.
#
## Inputs:
# X: the original signal
# N: the decomposition level
# Name: the wavelet name to use
#
## Outputs:
# cD is a N-row matrix containing the detail coefficients up to N levels
# cA is the same for the approximations

# This code was developed by David Springer for comparison purposes in the
# paper:
# D. Springer et al., "Logistic Regression-HSMM-based Heart Sound
# Segmentation," IEEE Trans. Biomed. Eng., In Press, 2015.
#
## Copyright (C) 2016  David Springer
# dave.springer@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
import pywt
from pywt import wavedec
import pyyawt.dwt1d
def wkeep1(x,l):
    start = int(np.floor((len(x) - l) / 2))
    c = x[start:start + l-1]
    return c
def getDWT(X,N,Name):
    #No DWT available for Morlet - therefore perform CWT:
    if (str(Name) == str('morl')):
        c = pywt.cwt(X,np.arange(1,N+1),'morl')
        cD = c
        cA = c
    else:
        #Preform wavelet decomposition
        X = np.squeeze(X)
        coeffs_1 = wavedec(X, Name, level=1,mode= 'symmetric')
        coeffs_2 = wavedec(X, Name, level=2,mode= 'symmetric')
        coeffs_3 = wavedec(X, Name, level=3,mode= 'symmetric')
        pyyawt.dwtmode("asymh")
        c,l = pyyawt.dwt1d.wavedec(X,N,Name)

        A1, D1 = coeffs_1
        A2, D2, D1 = coeffs_2
        A3, D3, D2, D1 = coeffs_3
        AppCoeff = (A1,A2,A3)

        # c= np.concatenate((A3, D3, D2, D1),axis = 0)
        # l = np.array([len(A3),len(D3),len(D2),len(D1)])
        #Reorder the details based on the structure of the wavelet
#decomposition (see help in wavedec.m)
        len_ = len(X)
        cD = np.zeros((N,len_))
        for k in np.arange(1,N+1).reshape(-1):
            d = pyyawt.dwt1d.detcoef(c,l,k)
            d = np.transpose(d)
            d= np.tile(d, (2 ** k,1))
            # d = d[np.ones((1,2 ** k)),:]
            cD[k-1, :] = pyyawt.utility.wkeep(np.transpose(d), [1, len_])

            # cD[k,:] = wkeep1(np.transpose(d),len_)
        cD = cD.flatten('F')
        #Space cD according to spacing of floating point numbers:
        I = np.where(np.abs(cD) < np.sqrt(np.finfo('float').eps))
        I = I[0]
        I_reshape = np.reshape(I,(I.shape[0],1))
        # I_reshape = I.shape[0]
        cD[I_reshape] = np.zeros((I.shape[0],1))
        cD = np.reshape(cD,(N,len_))
        # cD = wcodemat(cfd,nbcol,'row');
        #Reorder the approximations based on the structure of the wavelet
#decomposition (see help in wavedec.m)
        len_ = len(X)
        cA = np.zeros((N,len_))
        for k in np.arange(1,N+1).reshape(-1):
            #a = pyyawt.dwt1d.appcoef(c,l,Name,k)
            a = AppCoeff[k-1]
            a = np.transpose(a)
            a = np.tile(a, (2 ** k,1))
            cA[k-1,:] = pyyawt.utility.wkeep(np.transpose(a), [1, len_])
        cA = cA.flatten('F')
        I = np.where(np.abs(cA) < np.sqrt(np.finfo('float').eps))
        I = I[0]
        I_reshape = np.reshape(I,(I.shape[0],1))
        cA[I_reshape] = np.zeros((I.shape[0],1))
        cA = np.reshape(cA,(N,len_))
    
    return cD,cA