# function [heartRate systolicTimeInterval] = getHeartRateSchmidt(audio_data, Fs, figures)
#
# Derive the heart rate and the sytolic time interval from a PCG recording.
# This is used in the duration-dependant HMM-based segmentation of the PCG
# recording.
#
# This method is based on analysis of the autocorrelation function, and the
# positions of the peaks therein.
#
# This code is derived from the paper:
# S. E. Schmidt et al., "Segmentation of heart sound recordings by a
# duration-dependent hidden Markov model," Physiol. Meas., vol. 31,
# no. 4, pp. 513-29, Apr. 2010.
#
# Developed by David Springer for comparison purposes in the paper:
# D. Springer et al., "Logistic Regression-HSMM-based Heart Sound
# Segmentation," IEEE Trans. Biomed. Eng., In Press, 2015.
#
## INPUTS:
# audio_data: The raw audio data from the PCG recording
# Fs: the sampling frequency of the audio recording
# figures: optional boolean to display figures
#
## OUTPUTS:
# heartRate: the heart rate of the PCG in beats per minute
# systolicTimeInterval: the duration of systole, as derived from the
# autocorrelation function, in seconds
#
## Copyright (C) 2016  David Springer
# dave.springer@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from matplotlib.pyplot import xcorr

from butterworth_low_pass_filter import butterworth_low_pass_filter
from butterworth_high_pass_filter import butterworth_high_pass_filter
from schmidt_spike_removal import schmidt_spike_removal
from Homomorphic_Envelope_with_Hilbert import Homomorphic_Envelope_with_Hilbert
import numpy as np
import scipy
def getHeartRateSchmidt(audio_data,Fs):
    ## Get heatrate:
# From Schmidt:
# "The duration of the heart cycle is estimated as the time from lag zero
# to the highest peaks between 500 and 2000 ms in the resulting
# autocorrelation"
# This is performed after filtering and spike removal:
    
    ## 25-400Hz 4th order Butterworth band pass
    audio_data = np.round(audio_data,4)
    audio_data = butterworth_low_pass_filter(audio_data,2,400,Fs,False)
    audio_data = np.round(audio_data,4)
    audio_data = butterworth_high_pass_filter(audio_data,2,25,Fs)
    audio_data = np.round(audio_data,4)
    ## Spike removal from the original paper:
    audio_data = schmidt_spike_removal(audio_data,Fs)
    ## Find the homomorphic envelope
    homomorphic_envelope = Homomorphic_Envelope_with_Hilbert(audio_data,Fs)
    homomorphic_envelope = np.round(homomorphic_envelope,4)
    ## Find the autocorrelation:
    y = homomorphic_envelope - np.mean(homomorphic_envelope)
    cxx0 = np.sum(np.abs(y)** 2, axis = 0)
    cyy0 = np.sum(abs(y)** 2, axis = 0)
    scaleCoeffCross = np.sqrt(cxx0 * cyy0)

    c= scipy.signal.correlate(y, y, mode='full')
    c = np.divide(c,scaleCoeffCross)

    # timeseries = y
    # mean = np.mean(timeseries)
    # timeseries -= np.mean(timeseries)
    # autocorr_f = np.correlate(timeseries, timeseries, mode='full')
    # c = autocorr_f[autocorr_f.size / 2:] / autocorr_f[autocorr_f.size / 2]
    # iact.append(sum(autocorr_f[autocorr_f.size / 2:] / autocorr_f[autocorr_f.size / 2]))
    # c= np.corrcoef(y)# c = xcorr(y,'coeff')
    signal_autocorrelation = c[np.arange(len(homomorphic_envelope) ,len(c))]
    signal_autocorrelation = np.round(signal_autocorrelation,4)
    min_index = 0.5 * Fs
    max_index = 2 * Fs
    corr_segment = signal_autocorrelation[np.arange(int(min_index), int(max_index))]
    index = np.argmax(corr_segment)
    true_index = index + min_index - 1
    heartRate = 60 / (true_index / Fs)
    ## Find the systolic time interval:
# From Schmidt: "The systolic duration is defined as the time from lag zero
# to the highest peak in the interval between 200 ms and half of the heart
# cycle duration"
    
    max_sys_duration = np.round(((60 / heartRate) * Fs) / 2)
    min_sys_duration = np.round(0.2 * Fs)
    pos = np.argmax(signal_autocorrelation[np.arange(int(min_sys_duration),int(max_sys_duration+1))])
    systolicTimeInterval = (min_sys_duration + pos) / Fs
    return heartRate,systolicTimeInterval