# function [PCG_Features, featuresFs] = getSpringerPCGFeatures(audio_data, Fs, figures)
#
# Get the features used in the Springer segmentation algorithm. These
# features include:
# -The homomorphic envelope (as performed in Schmidt et al's paper)
# -The Hilbert envelope
# -A wavelet-based feature
# -A PSD-based feature
# This function was developed for use in the paper:
# D. Springer et al., "Logistic Regression-HSMM-based Heart Sound
# Segmentation," IEEE Trans. Biomed. Eng., In Press, 2015.
#
## INPUTS:
# audio_data: array of data from which to extract features
# Fs: the sampling frequency of the audio data
# figures (optional): boolean variable dictating the display of figures
#
## OUTPUTS:
# PCG_Features: array of derived features
# featuresFs: the sampling frequency of the derived features. This is set
# in default_Springer_HSMM_options.m
#
## Copyright (C) 2016  David Springer
# dave.springer@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
from default_Springer_HSMM_options import default_Springer_HSMM_options
from butterworth_low_pass_filter import butterworth_low_pass_filter
from butterworth_high_pass_filter import butterworth_high_pass_filter
from schmidt_spike_removal import schmidt_spike_removal
from Homomorphic_Envelope_with_Hilbert import Homomorphic_Envelope_with_Hilbert
from normalise_signal import normalise_signal
from Hilbert_Envelope import Hilbert_Envelope
from get_PSD_feature_Springer_HMM import get_PSD_feature_Springer_HMM
from getDWT import getDWT
import numpy as np
import scipy.signal
    
def getSpringerPCGFeatures(audio_data,Fs):
    # function PCG_Features = getSpringerPCGFeatures(audio, Fs)
# Get the features used in the Springer segmentation algorithm.
    
    springer_options = default_Springer_HSMM_options()
    # Check to see if the Wavelet toolbox is available on the machine:
    include_wavelet = springer_options['include_wavelet_feature']
    featuresFs = springer_options['audio_segmentation_Fs']
    
    ## 25-400Hz 4th order Butterworth band pass
    audio_data = butterworth_low_pass_filter(audio_data,2,400,Fs,False)
    audio_data = butterworth_high_pass_filter(audio_data,2,25,Fs)
    ## Spike removal from the original paper:
    audio_data = schmidt_spike_removal(audio_data,Fs)
    ## Find the homomorphic envelope
    homomorphic_envelope = Homomorphic_Envelope_with_Hilbert(audio_data,Fs)
    # Downsample the envelope:
    downsampled_homomorphic_envelope = scipy.signal.resample_poly(homomorphic_envelope, featuresFs, Fs)

    # normalise the envelope:
    downsampled_homomorphic_envelope = normalise_signal(downsampled_homomorphic_envelope)
    ## Hilbert Envelope
    hilbert_envelope = Hilbert_Envelope(audio_data,Fs)
    downsampled_hilbert_envelope = scipy.signal.resample_poly(hilbert_envelope,featuresFs, Fs)
    downsampled_hilbert_envelope = normalise_signal(downsampled_hilbert_envelope)
    ## Power spectral density feature:
    
    psd = np.transpose(get_PSD_feature_Springer_HMM(audio_data,Fs,40,60))
    psd = scipy.signal.resample(psd,len(downsampled_homomorphic_envelope))
    psd = normalise_signal(psd)
    ## Wavelet features:
    
    if (include_wavelet):
        wavelet_level = 3
        wavelet_name = 'bior3.9'
        # Audio needs to be longer than 1 second for getDWT to work:
        if (len(audio_data) < Fs * 1.025):
            audio_data = np.array([[audio_data],[np.zeros((np.round(0.025 * Fs),1))]])
        cD,cA = getDWT(audio_data,wavelet_level,wavelet_name)
        wavelet_feature = np.abs(cD[wavelet_level-1,:])
        wavelet_feature = wavelet_feature[np.arange(0,len(homomorphic_envelope))]
        downsampled_wavelet = scipy.signal.resample_poly(wavelet_feature,featuresFs , Fs)

        downsampled_wavelet = np.transpose(normalise_signal(downsampled_wavelet))
    
    ##
    psd = np.reshape(psd, (-1, 1))
    downsampled_wavelet = np.reshape(downsampled_wavelet, (-1, 1))
    if (include_wavelet):
        PCG_Features =  np.column_stack((downsampled_homomorphic_envelope,downsampled_hilbert_envelope,psd,downsampled_wavelet))
    else:
        PCG_Features =  np.column_stack((downsampled_homomorphic_envelope,downsampled_hilbert_envelope,psd))
    
    ## Plotting figures
    return PCG_Features,featuresFs