# function [despiked_signal] = schmidt_spike_removal(original_signal, fs)
#
# This function removes the spikes in a signal as done by Schmidt et al in
# the paper:
# Schmidt, S. E., Holst-Hansen, C., Graff, C., Toft, E., & Struijk, J. J.
# (2010). Segmentation of heart sound recordings by a duration-dependent
# hidden Markov model. Physiological Measurement, 31(4), 513-29.
#
# The spike removal process works as follows:
# (1) The recording is divided into 500 ms windows.
# (2) The maximum absolute amplitude (MAA) in each window is found.
# (3) If at least one MAA exceeds three times the median value of the MAA's,
# the following steps were carried out. If not continue to point 4.
#   (a) The window with the highest MAA was chosen.
#   (b) In the chosen window, the location of the MAA point was identified as the top of the noise spike.
#   (c) The beginning of the noise spike was defined as the last zero-crossing point before theMAA point.
#   (d) The end of the spike was defined as the first zero-crossing point after the maximum point.
#   (e) The defined noise spike was replaced by zeroes.
#   (f) Resume at step 2.
# (4) Procedure completed.
#
## Inputs:
# original_signal: The original (1D) audio signal array
# fs: the sampling frequency (Hz)
#
## Outputs:
# despiked_signal: the audio signal with any spikes removed.
#
# This code is derived from the paper:
# S. E. Schmidt et al., "Segmentation of heart sound recordings by a
# duration-dependent hidden Markov model," Physiol. Meas., vol. 31,
# no. 4, pp. 513-29, Apr. 2010.
#
# Developed by David Springer for comparison purposes in the paper:
# D. Springer et al., ?Logistic Regression-HSMM-based Heart Sound
# Segmentation,? IEEE Trans. Biomed. Eng., In Press, 2015.
#
## Copyright (C) 2016  David Springer
# dave.springer@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import numpy as np
    
def schmidt_spike_removal(original_signal,fs):
    ## Find the window size
# (500 ms)
    windowsize = int(np.round(fs / 2))
    ## Find any samples outside of a integer number of windows:
    trailingsamples = int(np.mod(len(original_signal),windowsize))
    ## Reshape the signal into a number of windows:
    temp = original_signal[np.arange(0, len(original_signal) - trailingsamples)]
    sampleframes = np.reshape(temp,(windowsize,int(len(temp)/windowsize)),order = 'F')
    ## Find the MAAs:
    MAAs = np.max(np.abs(sampleframes),axis=0)
    # While there are still samples greater than 3* the median value of the
# MAAs, then remove those spikes:
    while (not len(np.where((MAAs > np.median(MAAs) * 3))[0])==0 ):
        # print(np.median(MAAs) * 3)
        # print(len(np.where((MAAs > np.median(MAAs) * 3))[0]))
        #Find the window with the max MAA:
        window_num = np.argmax(MAAs)
        val = np.max(MAAs)
        # val,window_num = np.amax(MAAs)
        if (np.asarray(window_num).size > 1):
            window_num = window_num[0]
        #Find the postion of the spike within that window:
        # val,spike_position = np.amax(np.abs(sampleframes[:,window_num]))
        val= np.max(np.abs(sampleframes[:,window_num]))
        spike_position = np.argmax(np.abs(sampleframes[:,window_num]))
        if (np.asarray(spike_position).size > 1):
            spike_position = spike_position[0]
        # print(spike_position)
        # Finding zero crossings (where there may not be actual 0 values, just a change from positive to negative):
        zero_crossings = np.abs(np.diff(np.sign(sampleframes[:,window_num]))) > 1
        zero_crossings = zero_crossings*1
        zero_crossings = np.hstack((zero_crossings, 0))
        # zero_crossings = np.array([[np.abs(np.diff(np.sign(sampleframes[:,window_num]))) > 1],[0]])
        #Find the start of the spike, finding the last zero crossing before
#spike position. If that is empty, take the start of the window:
        temp = np.where(zero_crossings[np.arange(0, spike_position)] == 1)[0]

        if len(temp)==0:
            # print('Hi')
            spike_start = 0
        else:
            spike_start = np.max(np.array([0, temp[len(temp) - 1]]))
        #     spike_start = 0
        #Find the end of the spike, finding the first zero crossing after
#spike position. If that is empty, take the end of the window:
        if spike_position!=0:
            zero_crossings[np.arange(0,spike_position)] = 0
        else:
            zero_crossings[0] = 0
        temp = np.where(zero_crossings== 1)[0]
        if len(temp)==0:
            spike_end = windowsize-1
        else:
            spike_end = np.min(np.array([windowsize-1,temp[0]]))

        # spike_end = np.amin(np.array([(np.where(zero_crossings,1,'first')[0]),windowsize]))
        #Set to Zero
        sampleframes[np.arange(spike_start,spike_end+1),window_num] = 0.0001
        #Recaclulate MAAs
        MAAs = np.amax(np.abs(sampleframes),axis =0)

    
    despiked_signal = np.reshape(sampleframes,(-1,1), order = 'F')
    # Training_subset_reshaped= np.reshape(original_signal[np.arange(len(despiked_signal), len(original_signal))],(len(original_signal[np.arange(len(despiked_signal), len(original_signal))]),1))
    # Add the trailing samples back to the signal:
    temp_array = original_signal[np.arange(len(despiked_signal),len(original_signal))].reshape(-1,1)
    despiked_signal = np.vstack((despiked_signal,temp_array))
    return despiked_signal
