# Copyright 2022, TATA Consultancy Services. All rights reserved.
#!/usr/bin/env python
seed = 42
env={'PYTHONHASHSEED': seed}
import tensorflow
import random
tensorflow.keras.utils.set_random_seed(seed)
tensorflow.config.experimental.enable_op_determinism()
tensorflow.compat.v1.set_random_seed(seed)
tensorflow.random.set_seed(seed)
random.seed(seed)

from tensorflow.keras.optimizers import Adam, Nadam
from tensorflow.keras.callbacks import EarlyStopping, ModelCheckpoint
from tensorflow.keras.models import load_model
#
import numpy as np, scipy as sp, pandas as pd, os, sys, joblib, pickle
np.random.seed(seed)
import matplotlib.pyplot as pyplot
import cv2
import gc
from collections import Counter
import librosa
import librosa.display
import speechpy
#
from sklearn.utils import class_weight
from sklearn.utils.class_weight import compute_class_weight
from sklearn.model_selection import train_test_split
from sklearn.preprocessing import RobustScaler
import scipy.stats
from scipy.signal import resample_poly
from scipy.signal import spectrogram
from scipy.interpolate import griddata
from scipy.signal.windows import hann, kaiser, flattop
#
from Parameter import Parameter_dictionary
from CNNet_spectrogram import CNNet
from helper_code import *
from main_segmentation import pcg_segmentation
from pcg_noise_cleaning import test_get_clean_complete_states
from default_Springer_HSMM_options import default_Springer_HSMM_options
from butterworth_low_pass_filter import butterworth_low_pass_filter


################################################################################
#
# Required functions. Edit these functions to add your code, but do not change the arguments.
#
################################################################################

# Train your model.
def train_challenge_model(data_folder, model_folder, verbose):
    # Find data files.
    if verbose >= 1:
        print('Finding data files...')

    # Find the patient data files.
    patient_files = find_patient_files(data_folder)
    num_patient_files = len(patient_files)

    if num_patient_files==0:
        raise Exception('No data was provided.')

    # Create a folder for the model if it does not already exist.
    os.makedirs(model_folder, exist_ok=True)

    # Extract the features and labels.
    if verbose >= 1:
        print('Extracting features and labels from the Challenge data...')

    Param = Parameter_dictionary()

    # ====== Train Val Split=========================================#
    train_headers, valid_headers = train_test_split(patient_files, test_size=0.1, random_state=1)

    X_train, y_train_murmur, y_train_outcome, y_train_murmur_categ, y_train_outcome_categ = full_data_feat(data_folder, train_headers, Param)
    X_valid, y_valid_murmur, y_valid_outcome,  y_valid_murmur_categ, y_valid_outcome_categ = full_data_feat(data_folder, valid_headers, Param)

    print("Original murmur distribution", Counter(np.squeeze(y_train_murmur_categ)))
    print("Original outcome distribution", Counter(np.squeeze(y_train_outcome_categ)))

    # Train the model.
    if verbose >= 1:
        print('Training model...')
        
    X_train = np.ndarray.astype(X_train, np.float32)

    trainNetwork_murmur(X_train, y_train_murmur_categ, X_valid, y_valid_murmur_categ, model_folder, Param, "murmur")
    trainNetwork_murmur(X_train, y_train_outcome_categ, X_valid, y_valid_outcome_categ, model_folder, Param, "outcome")

    if verbose >= 1:
        print('Done.')


def get_segment(s1, cyc, current_recordings, frequency):
    seg_start = int(np.round(frequency * (s1[cyc, 0])))
    seg_end = int(np.round(frequency * (s1[cyc, 1])))
    segment = current_recordings[seg_start:seg_end]
    return segment


def remove_noisy_incomplete_cycle(sound_loc):
    arr = sound_loc[:, 2]
    pattern = [1, 2, 3, 4]
    # ---------------------------------
    pattern_length = len(pattern)
    potential_patterns = np.where(arr == pattern[0])[0]
    # ---------------------------------
    pattern_start_index = []
    for p in potential_patterns:
        check = arr[p:p + pattern_length]
        if np.all(check == pattern):
            pattern_start_index.append(p)

    pattern_start_index = np.array(pattern_start_index)
    # ---------------------------------
    a1 = np.zeros((len(pattern_start_index), pattern_length), dtype=np.int)
    for i in range(pattern_length):
        a1[:, i] = pattern_start_index + i
    index = np.hstack((a1))
    # ---------------------------------
    sound_loc_cleaned = sound_loc[index]

    return sound_loc_cleaned

def full_data_feat(data_folder, patient_files, Param ):
    num_of_segments_in_order = Param.get('num_of_segments_in_order')

    num_patient_files = len(patient_files)

    winsize = Param.get('winsize')
    overlap_window = Param.get('overlap_window')

    murmur_classes = Param.get('murmur_classes')
    outcome_classes = Param.get('outcome_classes')

    num_murmur_classes = len(murmur_classes)
    num_outcome_classes = len(outcome_classes)

    features = list()
    mur_labels = list()
    out_labels = list()
    files = list()

    mur_categ = list()
    out_categ = list()

    for i in range(num_patient_files):

        print('    {}/{}...'.format(i + 1, num_patient_files))

        # tsv file for '50782.txt' is not appropriate
        if os.path.basename(patient_files[i]) == '50782.txt':
            continue;

        # Load the current patient data
        current_patient_data = load_patient_data(patient_files[i])

        # Extract labels and use one-hot encoding.
        current_murmur = np.zeros(num_murmur_classes, dtype=int)
        murmur = get_murmur(current_patient_data)

        # Excluding unknown cases from training
        if murmur == 'Unknown':
            print("unknown")
            continue;

        if murmur in murmur_classes:
            j = murmur_classes.index(murmur)
            current_murmur[j] = 1

        if murmur == 'Present':
            murmur_binary = 1
        elif murmur == 'Absent':
            murmur_binary = 0

        current_outcome = np.zeros(num_outcome_classes, dtype=int)
        outcome = get_outcome(current_patient_data)
        if outcome in outcome_classes:
            j = outcome_classes.index(outcome)
            current_outcome[j] = 1

        if outcome == 'Abnormal':
            outcome_binary = 1
        elif outcome == 'Normal':
            outcome_binary = 0

        # =========================================
        num_locations = get_num_locations(current_patient_data)
        recording_information = current_patient_data.split('\n')[1:num_locations + 1]
        Sxx_all_windows_all_locations = []
        for l in range(num_locations):
            # print('Location', l)
            entries = recording_information[l].split(' ')
            recording_file = entries[2]
            filename = os.path.join(data_folder, recording_file)
            # Load recordings
            current_recordings, frequency = load_wav_file(filename)
            # ---Low pass filter-----
            current_recordings = butterworth_low_pass_filter(current_recordings, 2, 500, frequency, False)
            #------- Resampling and Normalization------------------------
            # raw_pcg_resamp will be fed into spectrogram
            # raw_pcg_norm_resamp is used to find the states
            raw_pcg_resamp, raw_pcg_norm_resamp, fs = preproc_test(current_recordings, frequency)
            # #---- sound states from segmentation code-------------#
            sound_loc = test_get_clean_complete_states(raw_pcg_norm_resamp, raw_pcg_resamp, fs, num_of_segments_in_order)
            # # -----------------------------------------------------
            # Get 2 cardiac cycles as one window and get spectrogram
            if sound_loc.shape[0] >= num_of_segments_in_order:
                Sxx_all_windows = get_Sxx_all_windows(sound_loc, raw_pcg_resamp, fs, winsize, overlap_window)
            else:
                Sxx_all_windows = 0
            # ----Sxx_all_windows is zero where there is only one complete cardiac cycle in sound loc
            if Sxx_all_windows == 0:
                print("Less than 2 complete cardiac cycle")
                continue;
            # --------------------------------------------------------
            Sxx_all_windows_all_locations.extend(Sxx_all_windows)
        # ----------------------------------------------------------------------------
        try:
            Sxx_all_windows_all_locations_numpy = np.vstack(Sxx_all_windows_all_locations)
        except:
            Sxx_all_windows_all_locations_numpy = np.random.rand(1, 100, 100)
        # ================================================

        files.append(os.path.basename(patient_files[i]))
        features.append(Sxx_all_windows_all_locations_numpy)
        mur_labels.append(np.tile(current_murmur, (Sxx_all_windows_all_locations_numpy.shape[0], 1)))
        out_labels.append(np.tile(current_outcome, (Sxx_all_windows_all_locations_numpy.shape[0], 1)))
        # =====================================
        mur_categ.append(np.tile(murmur_binary, (Sxx_all_windows_all_locations_numpy.shape[0], 1)))
        out_categ.append(np.tile(outcome_binary, (Sxx_all_windows_all_locations_numpy.shape[0], 1)))


    features = np.vstack(features)
    murmur_labels = np.vstack(mur_labels)
    outcome_labels = np.vstack(out_labels)
    mur_categ = np.vstack(mur_categ)
    out_categ = np.vstack(out_categ)

    return features, murmur_labels, outcome_labels, mur_categ, out_categ

def preproc_train(current_recordings, frequency):
    springer_options = default_Springer_HSMM_options()
    raw_pcg = current_recordings.astype(np.float32)
    raw_pcg_norm = get_norm(raw_pcg)
    fs = int(springer_options['audio_Fs'])

    # non-Normalized data is send to train
    raw_pcg_resamp = resample_poly(raw_pcg, fs, frequency)

    # raw_pcg_resamp will be fed into spectrogram
    # raw_pcg_norm_resamp is used to find the states
    return raw_pcg_resamp, fs

def preproc_test(current_recordings, frequency):
    springer_options = default_Springer_HSMM_options()
    raw_pcg = current_recordings.astype(np.float32)
    raw_pcg_norm = get_norm(raw_pcg)
    fs = int(springer_options['audio_Fs'])

    # non-Normalized data is send to train
    raw_pcg_resamp = resample_poly(raw_pcg, fs, frequency)

    # Normalization is not required in training code as we are taking the states from TSV files
    raw_pcg_norm_resamp = resample_poly(raw_pcg_norm, fs, frequency)

    # raw_pcg_resamp will be fed into spectrogram
    # raw_pcg_norm_resamp is used to find the states
    return raw_pcg_resamp, raw_pcg_norm_resamp, fs


def get_Sxx_all_windows(sound_loc, current_recordings, frequency, winsize, overlap_window):
    s1 = sound_loc[sound_loc[:, 2] == 1]
    systole = sound_loc[sound_loc[:, 2] == 2]
    s2 = sound_loc[sound_loc[:, 2] == 3]
    diastole = sound_loc[sound_loc[:, 2] == 4]
    #print('Segment shape', [s1.shape[0], systole.shape[0], s2.shape[0], diastole.shape[0]])
    # --------- Check for number of cycles
    if s1.shape[0] == systole.shape[0] == s2.shape[0] == diastole.shape[0]:
        num_cycles = s1.shape[0]
    else:
        num_cycles = min(s1.shape[0], systole.shape[0], s2.shape[0], diastole.shape[0])
        print("Incomplete cycle found")
    # print('A')
    if num_cycles > 1:  # 10 for 10 cycles
        cardiac_cycle_list = []
        for cyc in range(num_cycles):  # (s1_duration.shape[0]-1):
            #print('cyc', cyc)
            s1_segment = get_segment(s1, cyc, current_recordings, frequency)
            sy_segment = get_segment(systole, cyc, current_recordings, frequency)
            s2_segment = get_segment(s2, cyc, current_recordings, frequency)
            di_segment = get_segment(diastole, cyc, current_recordings, frequency)
            cardiac_cycle = np.hstack((s1_segment, sy_segment, s2_segment, di_segment))
            cardiac_cycle_list.append(cardiac_cycle)
            # print('A')
        # -----------------------------------------------------------------------

        WindowCount = int(np.floor((num_cycles - overlap_window) / (winsize - overlap_window)))

        Sxx_all_windows = []
        for i in range(WindowCount):
            a = i * (winsize - overlap_window)
            b = winsize + a
            #print('a,b: ', a, b)
            data_window = np.hstack((cardiac_cycle_list[a:b]))
            # augmentation
            #augmented_data_window = data_window + np.random.normal(0,np.std(data_window),len(data_window))
            Sxx = Generate_Spectrogram(data_window, frequency)
            #Sxx_aug = Generate_Spectrogram(augmented_data_window, frequency)
            Sxx_all_windows.append(Sxx)
            #Sxx_all_windows.append(Sxx_aug)
        return Sxx_all_windows
    else:
        Sxx_all_windows = 0
        return Sxx_all_windows


def Generate_Spectrogram(Overall_Max_SNR_window, fs):
    spec_fs = fs
    resampled_data = resample_poly(Overall_Max_SNR_window, spec_fs, fs)

    n_fft = 512
    S = librosa.feature.melspectrogram(y=resampled_data, sr=spec_fs, n_fft=n_fft, hop_length=int(0.01 * n_fft),
                                       center=False,
                                       power=2)
    mel_sgram = librosa.power_to_db(S, ref=np.max)
    # Mean and Variance Normalization
    mel_sgram = speechpy.processing.cmvn(
        mel_sgram,
        variance_normalization=True)

    # spectrogram_reshaped = mel_scale_sgram
    final_spec = cv2.resize(mel_sgram, (100, 100))
    spectrogram_reshaped = final_spec.reshape(1, final_spec.shape[0], final_spec.shape[1])
    return spectrogram_reshaped


def trainNetwork_murmur(X_train, y_train, X_valid, y_valid, model_folder, Param, netwrok_type):

    learning_rate = Param.get("learning_rate")

    nrw = X_train.shape[1]
    ncl = X_train.shape[2]

    if netwrok_type == "murmur":
        classes = Param.get("murmur_classes")
    else:
        classes = Param.get("outcome_classes")

    num_classes = len(classes)
    model = CNNet(nrw, ncl, num_classes)

    optimizer = Nadam(
        learning_rate=0.001,
        beta_1=0.9,
        beta_2=0.99,
        epsilon=1e-07,
        name='Nadam',
    )

    model.compile(loss='binary_crossentropy', optimizer=optimizer, metrics=['accuracy'],
                  weighted_metrics=['accuracy'])

    model.summary()

    filename = model_folder + "/model_" + netwrok_type + ".hdf5"

    checkpoint = ModelCheckpoint(filename, monitor='val_loss', verbose=0, save_best_only=True,
                                 save_weights_only=False,
                                 mode='auto', period=1)
    early = EarlyStopping(monitor='val_loss', min_delta=0, patience=7, verbose=0, mode='auto')

    reduce_lr = tensorflow.keras.callbacks.ReduceLROnPlateau(factor=0.1, patience=2, min_lr=learning_rate * 0.0001)

    print('xtrain, ytrain', X_train.shape, y_train.shape)
    print('xvalid, yvalid', X_valid.shape, y_valid.shape)

    y_train1 = np.squeeze(y_train)
    cls_weights = compute_class_weight(
        class_weight="balanced",
        classes=np.unique(y_train1),
        y=y_train1
    )
    cls_weight_dict = {0: cls_weights[0], 1: cls_weights[1]}
    val_sample_weights = class_weight.compute_sample_weight(cls_weight_dict, y_valid)

    history = model.fit([X_train], y_train, epochs=100, batch_size=32, verbose=1,
                        validation_data=([X_valid], y_valid, val_sample_weights),
                        callbacks=[checkpoint, reduce_lr, early])


# fig1 = pyplot.gcf()
    # # pyplot.yscale('log')
    # pyplot.title('Learning Curves')
    # pyplot.xlabel('Epoch')
    # pyplot.ylabel('Loss')
    # pyplot.plot(history.history['loss'], label='train')
    # pyplot.plot(history.history['val_loss'], label='val')
    # pyplot.legend()
    # fig1.savefig(model_folder + '/train_vs_eval_loss_'+ netwrok_type + '.png', dpi=300)
    # # pyplot.show()
    # pyplot.close()
    # fig2 = pyplot.gcf()
    # # pyplot.yscale('log')
    # pyplot.title('Learning Curves')
    # pyplot.xlabel('Epoch')
    # pyplot.ylabel('Accuracy')
    # pyplot.plot(history.history['accuracy'], label='train')
    # pyplot.plot(history.history['val_accuracy'], label='val')
    # pyplot.legend()
    # fig2.savefig(model_folder + '/train_vs_eval_accuracy_' + netwrok_type + '.png', dpi=300)
    # # pyplot.show()
    # pyplot.close()
    # # Save History
    # pd.DataFrame(history.history).to_csv(path_or_buf=model_folder + '/History_' + netwrok_type +'.csv')



# Load your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.
def load_challenge_model(model_folder, verbose):
    return model_folder

# Run your trained model. This function is *required*. You should edit this function to add your code, but do *not* change the
# arguments of this function.

def run_challenge_model(model, data, recordings, verbose):

    murmur_classifier = tensorflow.keras.models.load_model(model + "/model_murmur.hdf5" )
    outcome_classifier = tensorflow.keras.models.load_model(model + "/model_outcome.hdf5" )

    Param = Parameter_dictionary()
    murmur_classes = Param.get('murmur_classes')
    outcome_classes = Param.get('outcome_classes')
    OriginalSamplingFrequency = Param.get('OriginalSamplingFrequency')
    postprocess = Param.get('postprocess')
    pediction_decision = Param.get('pediction_decision')
    num_of_segments_in_order = Param.get('num_of_segments_in_order')
    winsize = Param.get('winsize')
    overlap_window = Param.get('overlap_window')

    num_locations = get_num_locations(data)

    spectrogram_reshaped = test_noise_cleaning_processing(num_locations, recordings, OriginalSamplingFrequency, num_of_segments_in_order, winsize, overlap_window)
    spectrogram_reshaped = np.ndarray.astype(spectrogram_reshaped, np.float32)

    murmur_classes, murmur_labels, murmur_probabilities = pred_postproc(murmur_classifier, spectrogram_reshaped,
                                                                        murmur_classes, pediction_decision)
    murmur_classes, murmur_labels, murmur_probabilities = postprocessing_unknown(murmur_probabilities, murmur_labels,
                                                                                 postprocess)
    outcome_classes, outcome_labels, outcome_probabilities = pred_postproc(outcome_classifier, spectrogram_reshaped,
                                                                           outcome_classes, pediction_decision)

    # Concatenate classes, labels, and probabilities.
    classes = murmur_classes + outcome_classes
    labels = np.concatenate((murmur_labels, outcome_labels))
    probabilities = np.concatenate((murmur_probabilities, outcome_probabilities))

    return classes, labels, probabilities

def get_norm(signal1):
    tim = ((1 - (-1)) * ((signal1 - np.min(signal1)) / (np.max(signal1) - np.min(signal1)))) + (-1)
    return tim


def test_noise_cleaning_processing(num_locations, current_recordings, frequency, num_of_segments_in_order, winsize, overlap_window):
    
    Sxx_all_windows_all_locations = []
    for l in range(num_locations):
        #print('Location', l)
        #------------------------------------------------
        current_recordings_location = current_recordings[l]
        # ---Low pass filter-----
        current_recordings_location = butterworth_low_pass_filter(current_recordings_location, 2, 500, frequency, False)
        #------- Resampling and Normalization------------------------
        # raw_pcg_resamp will be fed into spectrogram
        # raw_pcg_norm_resamp is used to find the states
        raw_pcg_resamp, raw_pcg_norm_resamp, fs = preproc_test(current_recordings_location, frequency)
        # #---- sound states from segmentation code-------------#
        sound_loc = test_get_clean_complete_states(raw_pcg_norm_resamp, raw_pcg_resamp, fs, num_of_segments_in_order)
        # # -----------------------------------------------------
        # Get 2 cardiac cycles as one window and get spectrogram
        if sound_loc.shape[0] >= num_of_segments_in_order:
            Sxx_all_windows = get_Sxx_all_windows(sound_loc, raw_pcg_resamp, fs, winsize, overlap_window)
        else:
            Sxx_all_windows = 0
        # ----Sxx_all_windows is zero where there is only one complete cardiac cycle in sound loc
        if Sxx_all_windows == 0:
            print("Less than 2 complete cardiac cycle")
            continue;
        #--------------------------------------------------------
        Sxx_all_windows_all_locations.extend(Sxx_all_windows)

    try:
        Sxx_all_windows_all_locations_numpy = np.vstack(Sxx_all_windows_all_locations)
    except:
        Sxx_all_windows_all_locations_numpy = np.random.rand(1, 100, 100)


    return Sxx_all_windows_all_locations_numpy


def pred_postproc(model, RecordingWindows, classes, pediction_decision):
    # print(RecordingWindows.shape)
    prob = model.predict(RecordingWindows)#, batch_size=10)
    probabilities = np.hstack((prob, 1 - prob))
    #probabilities = np.median(probabilities, axis=0)
    #labels = Pred_prob2Pred_class(probabilities)
    labels = np.apply_along_axis(Pred_prob2Pred_class, 1, probabilities)
    if pediction_decision == 1:
        # Majority Voting
        joint_lab = Pred_prob2Pred_class(np.sum(labels, axis=0))
        joint_prob = np.sum(labels, axis=0)/np.sum(labels)
    elif pediction_decision == 2:
        # Median
        joint_lab = Pred_prob2Pred_class(np.median(probabilities, axis=0))
        joint_prob = np.median(probabilities, axis=0)
    elif pediction_decision == 3:
        # Mean
        joint_lab = Pred_prob2Pred_class(np.mean(probabilities, axis=0))
        joint_prob = np.mean(probabilities, axis=0)
    elif pediction_decision == 4:
        # Threshold
        prob_morethan_point_zero_one = sum((prob>0.01)*1)
        final_prob = prob_morethan_point_zero_one/len(prob)

        if final_prob < 0.5:
            present = 1
            absent = 0
        else:
            absent = 1
            present = 0

        joint_prob = [present, absent]
        joint_lab = joint_prob

    return classes, joint_lab, joint_prob


def Pred_prob2Pred_class(Predicted_scores):
    Predicted_Class_indices = np.where(Predicted_scores == max(Predicted_scores))
    Predicted_Class_indices = np.array(Predicted_Class_indices)[0][0]
    Predicted_Classes = np.zeros((len(Predicted_scores),), dtype=int)
    Predicted_Classes[Predicted_Class_indices] = 1
    return Predicted_Classes

def postprocessing_unknown(probabilities, labels, postprocess):
    Present_prob = probabilities[0]
    Unknown_prob = 0.0
    Absent_prob = probabilities[1]
    murmur_probabilities = np.array([Present_prob, Unknown_prob, Absent_prob])

    if postprocess == 1:
        # --------------------------------
        Present_pred = 0
        Absent_pred = 0
        diff_prob = np.abs(Present_prob - Absent_prob)
        if diff_prob <= 0.35:
            Unknown_pred = 1
            murmur_labels = np.array([Present_pred, Unknown_pred, Absent_pred])
        else:
            Present_pred = labels[0]
            Unknown_pred = 0
            Absent_pred = labels[1]
            murmur_labels = np.array([Present_pred, Unknown_pred, Absent_pred])
            # -------------------------------------------
    else:
        Present_pred = labels[0]
        Unknown_pred = 0
        Absent_pred = labels[1]
        murmur_labels = np.array([Present_pred, Unknown_pred, Absent_pred])
    # -----------------------------------------------
    murmur_classes = ['Present', 'Unknown', 'Absent']
    return murmur_classes, murmur_labels, murmur_probabilities
