# function [delta, psi, qt] = viterbiDecodePCG_Springer(observation_sequence, pi_vector, b_matrix, total_obs_distribution, heartrate, systolic_time, Fs, figures)
#
# This function calculates the delta, psi and qt matrices associated with
# the Viterbi decoding algorithm from:
# L. R. Rabiner, "A tutorial on hidden Markov models and selected
# applications in speech recognition," Proc. IEEE, vol. 77, no. 2, pp.
# 257-286, Feb. 1989.
# using equations 32a - 35, and equations 68 - 69 to include duration
# dependancy of the states.
#
# This decoding is performed after the observation probabilities have been
# derived from the logistic regression model of Springer et al:
# D. Springer et al., "Logistic Regression-HSMM-based Heart Sound
# Segmentation," IEEE Trans. Biomed. Eng., In Press, 2015.
#
# Further, this function is extended to allow the duration distributions to extend
# past the beginning and end of the sequence. Without this, the label
# sequence has to start and stop with an "entire" state duration being
# fulfilled. This extension takes away that requirement, by allowing the
# duration distributions to extend past the beginning and end, but only
# considering the observations within the sequence for emission probability
# estimation. More detail can be found in the publication by Springer et
# al., mentioned above.
#
## Inputs:
# observation_sequence: The observed features
# pi_vector: the array of initial state probabilities, dervived from
# "trainSpringerSegmentationAlgorithm".
# b_matrix: the observation probabilities, dervived from
# "trainSpringerSegmentationAlgorithm".
# heartrate: the heart rate of the PCG, extracted using
# "getHeartRateSchmidt"
# systolic_time: the duration of systole, extracted using
# "getHeartRateSchmidt"
# Fs: the sampling frequency of the observation_sequence
# figures: optional boolean variable to show figures
#
## Outputs:
# logistic_regression_B_matrix:
# pi_vector:
# total_obs_distribution:
# As Springer et al's algorithm is a duration dependant HMM, there is no
# need to calculate the A_matrix, as the transition between states is only
# dependant on the state durations.
#
## Copyright (C) 2016  David Springer
# dave.springer@gmail.com
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import numpy as np
from scipy.io import loadmat
#from viterbi_Springer import viterbi_Springer
from get_duration_distributions import get_duration_distributions
from sklearn.linear_model import LogisticRegression
from default_Springer_HSMM_options import default_Springer_HSMM_options
from scipy.stats import multivariate_normal
def mnrval(beta,x):

    n = x.shape[0] #nrw
    p = x.shape[1] #ncl

    pstar = beta.shape[0]
    k = beta.shape[1] + 1
    eta = np.tile(beta[0,:],(n,1)) + np.dot(x , beta[1:pstar,:])# if matrix multiplication use np.dot(x,beta[1:pstar+1,:])
    max_eta = np.max(eta, axis=1).reshape(-1,1)
    pi = np.column_stack([np.exp(eta - max_eta), np.exp(-max_eta)])#rescale so max probability is 1
    pi = np.divide(pi, np.tile(np.sum(pi, axis = 1).reshape(-1,1), (1, k)))#renormalize for real probabilities
    return pi

def viterbiDecodePCG_Springer(observation_sequence ,pi_vector ,b_matrix,total_obs_distribution,heartrate,systolic_time ,Fs):
    ## Preliminary
    # observation_sequence = loadmat('pcg_features.mat')
    # observation_sequence = observation_sequence['observation_sequence']

    # psi_duration_matlab = loadmat('psi_duration_matlab.mat')
    # psi_duration_matlab = psi_duration_matlab['psi_duration_matlab']

    springer_options = default_Springer_HSMM_options()
    T = len(observation_sequence)
    N = 4
    
    # Setting the maximum duration of a single state. This is set to an entire
# heart cycle:
    max_duration_D = int(np.round((1 * (60 / heartrate)) * Fs))
    #Initialising the variables that are needed to find the optimal state path along
#the observation sequence.
#delta_t(j), as defined on page 264 of Rabiner, is the best score (highest
#probability) along a single path, at time t, which accounts for the first
#t observations and ends in State s_j. In this case, the length of the
#matrix is extended by max_duration_D samples, in order to allow the use
#of the extended Viterbi algortithm:
    delta = np.ones((T + max_duration_D - 1,N)) * - np.inf
    #The argument that maximises the transition between states (this is
#basically the previous state that had the highest transition probability
#to the current state) is tracked using the psi variable.
    psi = np.zeros((T + max_duration_D - 1,N))
    #An additional variable, that is not included on page 264 or Rabiner, is
#the state duration that maximises the delta variable. This is essential
#for the duration dependant HMM.
    psi_duration = np.zeros((T + max_duration_D - 1,N))
    ## Setting up observation probs
    observation_probs = np.zeros((T,N))
    for n in np.arange(0,N).reshape(-1):
        #MLR gives P(state|obs)
#Therefore, need Bayes to get P(o|state)
#P(o|state) = P(state|obs) * P(obs) / P(states)
#Where p(obs) is derived from a MVN distribution from all
#obserbations, and p(states) is taken from the pi_vector:

        # model = LogisticRegression(random_state=0, multi_class='multinomial', solver='newton-cg')
        # model.fit(b_matrix(n),observation_sequence[:,:])  # evaluating the score of the model
        pihat = mnrval(b_matrix[0][n],observation_sequence)
        #pihat = mnrval(cell2mat(b_matrix(n)),observation_sequence(:,:))
        for t in np.arange(0,T).reshape(-1):
            Po_correction = multivariate_normal.pdf(np.transpose(observation_sequence[t,:]).squeeze(),mean = np.transpose(total_obs_distribution[0][0]).squeeze(), cov = total_obs_distribution[1][0])
            # Po_correction =rv.pdf(observation_sequence[t,:])
            # Po_correction = multivariate_normal.pdf(np.transpose(observation_sequence[t,:]).reshape(1,-1),mean=np.transpose(total_obs_distribution[0][0]),cov= total_obs_distribution[1][0])
            #When saving the coefficients from the logistic
#regression, it orders them P(class 1) then P(class 2). When
#training, I label the classes as 0 and 1, so the
#correct probability would be pihat(2).
            observation_probs[t,n] = (pihat[t,1] * Po_correction) / pi_vector[0,n]
    # observation_probs = np.round(observation_probs,4)
    ## Setting up state duration probabilities, using Gaussian distributions:
    d_distributions,max_S1,min_S1,max_S2,min_S2,max_systole,min_systole,max_diastole,min_diastole = get_duration_distributions(heartrate,systolic_time)
    duration_probs = np.zeros((N,3 * Fs))
    duration_sum = np.zeros((N,1))
    for state_j in np.arange(0,N).reshape(-1):
        for d in np.arange(1,max_duration_D+1).reshape(-1):
            if (state_j == 0):
                duration_probs[state_j,d-1] = multivariate_normal.pdf(d,d_distributions[state_j,0],d_distributions[state_j,1])
                if (d < min_S1 or d > max_S1):
                    duration_probs[state_j,d-1] = np.finfo(float).tiny
            else:
                if (state_j == 2):
                    duration_probs[state_j,d-1] = multivariate_normal.pdf(d,d_distributions[state_j,0],d_distributions[state_j,1])
                    if (d < min_S2 or d > max_S2):
                        duration_probs[state_j,d-1] = np.finfo(float).tiny
                else:
                    if (state_j == 1):
                        duration_probs[state_j,d-1] = multivariate_normal.pdf(d,d_distributions[state_j,0],d_distributions[state_j,1])
                        if (d < min_systole or d > max_systole):
                            duration_probs[state_j,d-1] = np.finfo(float).tiny
                    else:
                        if (state_j == 3):
                            duration_probs[state_j,d-1] = multivariate_normal.pdf(d,d_distributions[state_j,0],d_distributions[state_j,1])
                            if (d < min_diastole or d > max_diastole):
                                duration_probs[state_j,d-1] = np.finfo(float).tiny
        duration_sum[state_j] = sum(duration_probs[state_j,:])
    # duration_probs = np.round(duration_probs,4)
    if (duration_probs.shape[1] > 3 * Fs):
        duration_probs = duration_probs[:,np.arange(0,3*Fs)]
    
    ## Perform the actual Viterbi Recursion:
    
    qt = np.zeros((1,len(delta)))
    ## Initialisation Step
    
    #Equation 32a and 69a, but leave out the probability of being in
#state i for only 1 sample, as the state could have started before time t =
#0.
    
    delta[0,:] = np.log(pi_vector) + np.log(observation_probs[0,:])
    
    #Equation 32b
    psi[0,:] = - 1
    # The state duration probabilities are now used.
#Change the a_matrix to have zeros along the diagonal, therefore, only
#relying on the duration probabilities and observation probabilities to
#influence change in states:
#This would only be valid in sequences where the transition between states
#follows a distinct order.
    a_matrix = np.array([[0,1,0,0],[0,0,1,0],[0,0,0,1],[1,0,0,0]])
    ## Run the core Viterbi algorith
    
    if (springer_options['use_mex']):
        ## Run Mex code
        Hello = 1 ##dummy line
# Ensure you have run the mex viterbi_PhysChallenge.c code on the
# native machine before running this:
        #delta,psi,psi_duration = viterbi_Springer(N,T,a_matrix,max_duration_D,delta,observation_probs,duration_probs,psi,duration_sum)
    else:
        ## Recursion
        ## The Extended Viterbi algorithm:
        #Equations 33a and 33b and 69a, b, c etc:
#again, ommitting the p(d), as state could have started before t = 1
        # This implementation extends the standard implementation of the
# duration-dependant Viterbi algorithm by allowing the durations to
# extend beyond the start and end of the time series, thereby allowing
# states to "start" and "stop" outside of the recorded signal. This
# addresses the issue of partial states at the beginning and end of the
# signal being labelled as the incorrect state. For instance, a
# short-duration diastole at the beginning of a signal looks a lot like
# systole, and can lead to labelling errors.
        max_delta_counter = np.zeros((130000,1))
        counter = int(0)
        # t spans input 2 to T + max_duration_D:
        for t in np.arange(1,T + max_duration_D-1).reshape(-1):
            for j in np.arange(0,N).reshape(-1):
                for d in np.arange(1,max_duration_D+1,1).reshape(-1):
                    #The start of the analysis window, which is the current time
#step, minus d (the time horizon we are currently looking back),
#plus 1. The analysis window can be seen to be starting one
#step back each time the variable d is increased.
# This is clamped to 1 if extending past the start of the
# record, and T-1 is extending past the end of the record:
                    start_t = t - d
                    if (start_t < 0):
                        start_t = 0
                    if (start_t > T-1 ):
                        start_t = T - 1
                    #The end of the analysis window, which is the current time
#step, unless the time has gone past T, the end of the record, in
#which case it is truncated to T. This allows the analysis
#window to extend past the end of the record, so that the
#timing durations of the states do not have to "end" at the end
#of the record.
                    end_t = t
                    if (t > T-1):
                        end_t = T-1
                    #Find the max_delta and index of that from the previous step
#and the transition to the current step:
#This is the first half of the expression of equation 33a from
#Rabiner:
                    max_index = np.argmax(delta[start_t,:] + np.transpose(np.log(a_matrix[:,j])),axis = 0)
                    max_delta = np.max(delta[start_t,:] + np.transpose(np.log(a_matrix[:,j])),axis = 0)

                    #Find the normalised probabilities of the observations over the
#analysis window:
                    probs = np.prod(observation_probs[np.arange(start_t,end_t+1),j])
                    #Find the normalised probabilities of the observations at only
#the time point at the start of the time window:
                    if (probs == 0):
                        probs = np.finfo(float).tiny
                    emission_probs = np.log(probs)
                    #Keep a running total of the emmission probabilities as the
#start point of the time window is moved back one step at a
#time. This is the probability of seeing all the observations
#in the analysis window in state j:
                    if (emission_probs == 0 or np.isnan(emission_probs)):
                        emission_probs = np.finfo(float).tiny
                    #Find the total probability of transitioning from the last
#state to this one, with the observations and being in the same
#state for the analysis window. This is the duration-dependant
#variation of equation 33a from Rabiner:
                    delta_temp = max_delta + (emission_probs) + np.log((duration_probs[j,d-1] / duration_sum[j]))
                    # max_delta_counter[counter, 0] = max_delta
                    # counter = counter + 1
                    # if counter == 13937:
                    #     print('got you!!')
                    # # if  t == 135 and j == 0:
                    # print('max_index',max_index,'emission_probs',emission_probs,'j',j,'d',d,'delta_temp',delta_temp)
                   #Unlike equation 33a from Rabiner, the maximum delta could come
#from multiple d values, or from multiple size of the analysis
#window. Therefore, only keep the maximum delta value over the
#entire analysis window:
#If this probability is greater than the last greatest,
#update the delta matrix and the time duration variable:
                    if (delta_temp > delta[t,j]):
                        delta[t,j] = delta_temp
                        psi[t,j] = max_index
                        psi_duration[t,j] = d
    
    ## Termination
    
    # For the extended case, need to find max prob after end of actual
# sequence:
    
    # Find just the delta after the end of the actual signal
    temp_delta = delta[np.arange(T,len(delta)),:]
    #Find the maximum value in this section, and which state it is in:
    idx = np.argmax(np.ndarray.flatten(temp_delta))
    pos,__ = np.unravel_index(idx,temp_delta.shape)
    # Change this position to the real position in delta matrix:
    pos = pos + T
    #1) Find the last most probable state
#2) From the psi matrix, find the most likely preceding state
#3) Find the duration of the last state from the psi_duration matrix
#4) From the onset to the offset of this state, set to the most likely state
#5) Repeat steps 2 - 5 until reached the beginning of the signal
    
    #The initial steps 1-4 are equation 34b in Rabiner. 1) finds P*, the most
#likely last state in the sequence, 2) finds the state that precedes the
#last most likely state, 3) finds the onset in time of the last state
#(included due to the duration-dependancy) and 4) sets the most likely last
#state to the q_t variable.
    
    #1)
    state = np.argmax(delta[pos,:].reshape(-1,1),axis=0)
    #2)
    offset = pos
    preceding_state = psi[offset,state]
    #3)
# state_duration = psi_duration(offset, state);
    onset = offset - psi_duration[offset,state] + 1
    #4)
    qt[0,np.arange(int(onset[0]),offset+1)] = state
    qt = qt.astype(int)
    #The state is then updated to the preceding state, found above, which must
#end when the last most likely state started in the observation sequence:
    state = int(preceding_state[0])
    count = 0
    onset = int(onset[0])
    offset = int(offset)
    #While the onset of the state is larger than the maximum duration
#specified:
    while (onset > 2):

        #2)
        offset = onset - 1
        #     offset_array(offset,1) = inf;
        preceding_state = int(psi[offset,state])
        #     offset_array(offset,2) = preceding_state;
        #3)
#     state_duration = psi_duration(offset, state);
        onset = offset - int(psi_duration[offset,state]) + 1
        #4)
#     offset_array(onset:offset,3) = state;
        if (onset < 2):
            onset = 0
        qt[0,np.arange(onset,offset+1)] = state
        state = preceding_state
        count = count + 1
        if (count > 1000):
            break

    
    qt = qt[0,np.arange(0,T)]
    return delta,psi,qt